import datetime

from . import TypeModel
from .events import Event
from .rooms import Room


class Booking(TypeModel):
    base_path = "/booking/"
    def __init__(
            self, 
            time_start: datetime.datetime,
            time_finish: datetime.datetime, 
            event_id: int | None = None,
            room_id: int | None = None,
            event: Event | None = None,
            room: Room | None = None,
            id: int | None = None,
            additional_data: dict | None = None):
        """
            Initialize a Booking object.

            :param time_start: The start time of the booking as a datetime object.
            :param time_finish: The end time of the booking as a datetime object.
            :param event_id: The ID of the associated event (optional).
            :param room_id: The ID of the associated room (optional).
            :param event: An Event object associated with the booking (optional).
            :param room: A Room object associated with the booking (optional).
            :param id: The ID of the booking (optional).
            :param additional_data: Additional data related to the booking (optional).

            :raises ValueError: If event and room instances are not provided and event_id or room_id is missing.
        """
        
        self.id = id
        if (not event or not room) and (not event_id or not room_id):
            raise ValueError("You should pass event and room instances or their IDs")
        self.event = event
        self.room = room
        self.event_id = event_id
        self.room_id = room_id
        self.time_start = time_start
        self.time_finish = time_finish
        self.additional_data = additional_data

    @classmethod
    def from_json(cls, booking: dict):
        """
            Create a Booking object from a JSON dictionary obtained from the API.

            :param booking: JSON dictionary representing a booking.

            :return: A Booking object.
        """
        return cls(
            id=booking["id"],
            event_id=booking["event_id"], 
            room_id=booking["room_id"], 
            time_start=datetime.datetime.fromisoformat(booking["time_start"]), 
            time_finish=datetime.datetime.fromisoformat(booking["time_finish"]),
            additional_data=booking["additional_data"])

    @property
    def body(self):
        """
            Get the data of the Booking object in a dictionary format for API requests.

            Note: This method may cause an error if the associated room or event is not saved in the database.

            :return: Dictionary representation of the Booking object.
        """
        return dict(
            time_start=self.time_start.isoformat(),
            time_finish=self.time_finish.isoformat(), 
            additional_data=self.additional_data,
            room_id=self.room.id if self.room else self.room_id,
            event_id=self.event.id if self.event else self.event_id)

    @property
    def params(self) -> dict:
        """
            Get query parameters for API requests (currently empty).

            :return: An empty dictionary.
        """
        return {}



from . import TypeModel


class Seat(TypeModel):
    base_path = "/seats/"

    def __init__(
            self,
            row: int,
            column: int,
            number: int,
            booked: bool = False,
            id: int | None = None,
            additional_data: dict | None = None):
        """
            Initialize a Seat object.

            :param row: The row number of the seat.
            :param column: The column number of the seat.
            :param number: The seat number.
            :param booked: Whether the seat is booked (default is False).
            :param id: The ID of the seat (optional).
            :param additional_data: Additional data related to the seat (optional).
        """
        self.id = id
        self.row = row
        self.column = column
        self.number = number
        self.booked = booked
        self.additional_data = additional_data

    @classmethod
    def from_json(cls, seat: dict):
        """
            Create a Seat object from a JSON dictionary obtained from the API.

            :param seat: JSON dictionary representing a seat.

            :return: A Seat object.
        """
        return cls(
            id=seat["id"],
            row=seat["row"], 
            column=seat["column"], 
            number=seat["number"], 
            booked=seat["booked"], 
            additional_data=seat["additional_data"]
        )

    @property
    def body(self) -> dict:
        """
            Get the data of the Seat object in a dictionary format for API requests.

            :return: Dictionary representation of the Seat object.
        """
        return dict(
            row=self.row,
            column=self.column, 
            number=self.number, 
            booked=self.booked, 
            additional_data=self.additional_data
        )

    @property
    def params(self) -> dict:
        """
            Get query parameters for API requests (currently empty).

            :return: An empty dictionary.
        """
        return {}



from . import TypeModel


class Event(TypeModel):
    base_path = "/events/"
    
    def __init__(
            self,
            title: str,
            id: int | None = None,
            additional_data: dict | None = None):
        """
            Initialize an Event object.

            :param title: The title of the event.
            :param id: The ID of the event (optional).
            :param additional_data: Additional data related to the event (optional).
        """
        self.id = id
        self.title = title
        self.additional_data = additional_data

    @classmethod
    def from_json(cls, event: dict):
        """
            Create an Event object from a JSON dictionary obtained from the API.

            :param event: JSON dictionary representing an event.

            :return: An Event object.
        """
        return cls(id=event["id"], title=event["title"], additional_data=event["additional_data"])

    @property
    def body(self) -> dict:
        """
            Get the data of the Event object in a dictionary format for API requests.

            :return: Dictionary representation of the Event object.
        """
        return dict(title=self.title, additional_data=self.additional_data)

    @property
    def params(self) -> dict:
        """
            Get query parameters for API requests (currently empty).

            :return: An empty dictionary.
        """
        return {}



from abc import ABC, abstractclassmethod, abstractproperty


class TypeModel(ABC):
    @abstractclassmethod
    def from_json(cls, object_json: dict):
        """
            Create an object of the implementing class from a JSON dictionary obtained from the API.

            :param object_json: JSON dictionary representing an object.

            :return: An instance of the implementing class.
        """
        raise NotImplementedError()
    
    @abstractproperty
    def body(self):
        """
            Get the data of the object in a dictionary format for API requests.

            :return: Dictionary representation of the object's data.
        """
        raise NotImplementedError()
    
    @abstractproperty
    def params(self):
        """
            Get query parameters for API requests.

            :return: Dictionary of query parameters.
        """
        raise NotImplementedError()



from book_system.types.booking import Booking
from book_system.types.events import Event
from book_system.types.seats import Seat
from book_system.types.rooms import Room
from book_system.sdk import BookSystemSDK



import requests
from typing import Literal

from .types import TypeModel
from .types.rooms import Room
from .types.events import Event
from .types.booking import Booking


class BookSystemSDK:
    def __init__(
            self,
            api_url: str,
            rooms: list[Room] | None = None,
            events: list[Event] | None = None,
            booking: list[Booking] | None = None):
        """
            Initialize the BookSystemSDK.

            :param api_url: The base URL for the API.
            :param rooms: A list of Room objects (optional).
            :param events: A list of Event objects (optional).
            :param booking: A list of Booking objects (optional).
        """
        self.api_url = api_url
        self._rooms = rooms
        self._events = events
        self._booking = booking
    
    @property
    def rooms(self) -> list[Room]:
        """
            Get a list of Room objects from the API.

            :return: A list of Room objects.
        """
        if not self._rooms:
            url = f"{self.api_url}/rooms/"
            self._rooms = [Room.from_json(room) for room in self._make_request(url=url, method="GET")]
        return self._rooms

    @property
    def events(self) -> list[Event]:
        """
            Get a list of Event objects from the API.

            :return: A list of Event objects.
        """
        if not self._events:
            url = f"{self.api_url}/events/"
            self._events = [Event.from_json(event) for event in self._make_request(url=url, method="GET")]
        return self._events
    
    @property
    def booking(self) -> list[Booking]:
        """
            Get a list of Booking objects from the API.

            :return: A list of Booking objects.
        """
        if not self._booking:
            url = f"{self.api_url}/booking/"
            self._booking = [Booking.from_json(booking) for booking in self._make_request(url=url, method="GET")]
        return self._booking

    def _make_request(
            self,
            url: str,
            method: Literal["GET", "POST", "PATCH", "DELETE"],
            body: dict | None = None,
            params: dict | None = None) -> dict:
        """
            Make an HTTP request to the API.

            :param url: The URL to make the request to.
            :param method: The HTTP method for the request.
            :param body: The request body data (optional).
            :param params: Query parameters for the request (optional).

            :return: The JSON response from the API.
        """
        response = requests.request(method=method, url=url, json=body, params=params)
        json = response.json() if response.status_code != 204 else None
        if response.status_code not in [200, 201, 204]:
            raise ValueError(json["detail"])
        return json
    
    def create(self, obj: TypeModel) -> TypeModel:
        """
            Create a new object on the API.

            :param obj: An object of a specific type to create.

            :return: The newly created object.
        """
        url = f"{self.api_url}{obj.base_path}"
        return obj.from_json(self._make_request(url=url, method="POST", body=obj.body, params=obj.params))
    
    def refresh(self, obj: TypeModel) -> TypeModel:
        """
            Refresh the data of an existing object on the API.

            :param obj: An object of a specific type to refresh.

            :return: The refreshed object.
        """
        url = f"{self.api_url}{obj.base_path}{obj.id}/"
        return obj.from_json(self._make_request(url=url, method="PATCH", body=obj.body))

    def delete(self, obj: TypeModel | list[TypeModel]) -> None:
        """
            Delete an object or a list of objects from the API.

            :param obj: An object or list of objects to delete.
        """
        url = f"{self.api_url}{obj.base_path}{obj.id}"
        self._make_request(url=url, method="DELETE")

    def get(self, model: TypeModel, by: TypeModel | None = None, by_id: int | None = None,  **kwargs) -> TypeModel:
        """
            Retrieve an object from the API.

            :param model: The type of object to retrieve.
            :param by: An optional filter object to narrow down the query.
            :param by_id: An optional ID to retrieve a specific object.
            :param kwargs: Additional query parameters.

            :return: The retrieved object or None if not found.
        """
        if not by and by_id:
            url = f"{self.api_url}{model.base_path}{by_id}/"
        elif by and by_id:
            url = f"{self.api_url}{by.base_path}{by_id}{model.base_path}"
        else:
            url = f"{self.api_url}{model.base_path}"
        json = self._make_request(url=url, method="GET", params=kwargs)
        if not isinstance(json, list):
            return model.from_json(json) if json else None
        return model.from_json(json[0]) if len(json) == 1 else [model.from_json(obj) for obj in json]



from setuptools import setup, find_packages


def get_readme():
    with open("README.md", "r") as file:
        return file.read()


setup(
  name='book-system-SDK',
  version='0.1.5',
  author='Artem Sydorenko',
  author_email='kradworkmail@gmail.com',
  description='SDK for interaction with the book-system API',
  long_description=get_readme(),
  long_description_content_type='text/markdown',
  packages=find_packages(),
  classifiers=[
    'Programming Language :: Python :: 3.11'
  ],
  python_requires='>=3.11'
)



from .seats import Seat
from . import TypeModel


class Room(TypeModel):
    base_path = "/rooms/"
    def __init__(
            self, 
            id: int | None = None,
            name: str | None = None, 
            seats: list[Seat] | None = None,
            autogenerate_seats: bool = False, 
            columns: int | None = None,
            rows: int | None = None):
        """
            Initialize a Room object.

            :param id: The ID of the room (optional).
            :param name: The name of the room (optional).
            :param seats: A list of Seat objects associated with the room (optional).
            :param autogenerate_seats: Whether to automatically generate seats (default is False).
            :param columns: The number of columns for seat generation (optional).
            :param rows: The number of rows for seat generation (optional).

            :raises ValueError: If autogenerate_seats is True but columns or rows are not provided.
        """
        
        if autogenerate_seats and (not columns or not rows):
            raise ValueError("If you want to use autogenerate you should pass the number of columns and rows")
        self.id = id
        self.name = name
        self.columns = columns
        self.rows = rows
        self.seats = seats
        self.autogenerate_seats = autogenerate_seats

    @classmethod
    def from_json(cls, room: dict):
        """
            Create a Room object from a JSON dictionary obtained from the API.

            :param room: JSON dictionary representing a room.

            :return: A Room object.
        """
        seats = room.get("seats")
        seats = [Seat.from_json({**seat}) for seat in seats] if seats else None
        return cls(id=room["id"], name=room["name"], seats=seats)

    @property
    def body(self) -> dict:
        """
            Get the data of the Room object in a dictionary format for API requests.

            :return: Dictionary representation of the Room object.
        """
        seats_dict = [seat.body for seat in self.seats] if self.seats else None
        return dict(
            name=self.name, 
            seats=seats_dict
        )

    @property
    def params(self) -> dict:
        """
            Get query parameters for API requests.

            :return: Dictionary with autogenerate, columns, and rows parameters.
        """
        return dict(
            autogenerate=self.autogenerate_seats, 
            columns=self.columns, 
            rows=self.rows
        )
